<?php

namespace Drupal\tempest_highlight\Plugin\Filter;

use Drupal\Component\Utility\Html;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Utility\Error as Error;
use Drupal\filter\Annotation\Filter;
use Drupal\filter\FilterProcessResult;
use Drupal\filter\Plugin\FilterBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Tempest\Highlight\Highlighter;
use Tempest\Highlight\Themes\InlineTheme;

/**
 * Provides a filter to highlight code using tempestphp/highlight.
 *
 * @Filter(
 *   id = "tempest_highlight",
 *   title = @Translation("Process code blocks with Tempest Highlight"),
 *   type = Drupal\filter\Plugin\FilterInterface::TYPE_TRANSFORM_REVERSIBLE
 * )
 */
class TempestHighlight extends FilterBase implements ContainerFactoryPluginInterface {

  protected FileSystemInterface $fileSystem;

  const THEMES_BASE_PATH = DRUPAL_ROOT . '/../vendor/tempest/highlight/src/Themes/Css';

  const DEFAULT_THEME = 'dark-plus';

  public function __construct(array $configuration, $plugin_id, $plugin_definition, ?FileSystemInterface $fileSystem) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->fileSystem = $fileSystem;
  }

  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('file_system')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function process($text, $langcode): FilterProcessResult {

    $result = new FilterProcessResult($text);
    if (stristr($text, '<code') === FALSE || stristr($text, 'language-') === FALSE) {
      return $result;
    }

    $theme = $this->settings['theme'] ?? self::DEFAULT_THEME;

    try {
      $dom = Html::load($text);
      $xpath = new \DOMXPath($dom);
      $highlighter = (new Highlighter(new InlineTheme(self::THEMES_BASE_PATH . '/' . $theme . '.css')));
    } catch (\Exception $e) {
      Error::logException(\Drupal::logger('tempest_highlight'), $e);
      return $result;
    }

    foreach ($xpath->query('//code[starts-with(@class, "language-")]') as $node) {
      if (!$language = $this->getCodeLanguage($node->getAttribute('class'))) {
        continue;
      }

      $codeContent = $node->nodeValue;
      $codeHighlighted = $highlighter->parse($codeContent, $language);
      $newNode = $dom->createDocumentFragment();
      $newNode->appendXML("<code>{$codeHighlighted}</code>");
      $node->parentNode->replaceChild($newNode, $node);
    }

    $result->setProcessedText(Html::serialize($dom));

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {

    $cssFiles = array_map(function ($file) {
      return str_replace('.css', '', $file->filename);
    }, $this->fileSystem->scanDirectory(self::THEMES_BASE_PATH, '/\.(css)$/'));

    ksort($cssFiles);

    $form['theme'] = [
      '#type' => 'select',
      '#title' => $this->t('Theme'),
      '#options' => array_combine($cssFiles, $cssFiles),
      '#default_value' => $this->settings['theme'] ?? self::DEFAULT_THEME,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function tips($long = FALSE) {
    return $this->t('Highlight code-blocks using Tempest Highlight');
  }

  private function getCodeLanguage($classes): ?string {
    foreach (explode(' ', $classes) as $class) {
      if (str_starts_with($class, 'language-')) {
        return substr($class, strlen('language-'));
      }
    }

    return NULL;
  }

}
